TAG_LIST_DELIMETER = ','
ACTION_DELIMETER = ':'
ANYONE_PRINCIPAL = 'any'
FULL_ACCESS_ACTION = 'all'
ROOT_PRINCIPAL = 'root'
VOID_PRINCIPAL = 'void'
VOID_RESOURCE = ''


class TagthException(Exception):
    pass


class TagthValidationError(TagthException):
    pass


def _normalize_principal(principal):
    def norm_item(item):
        tag = item.strip()

        if not tag:
            tag = VOID_PRINCIPAL

        if not tag.isidentifier():
            raise TagthValidationError(f'Special characters in principal tag: {tag}')

        return tag

    if not isinstance(principal, str):
        raise TagthValidationError(f'Bad principal {principal}')

    principal = principal.split(TAG_LIST_DELIMETER)
    return list(map(norm_item, principal))


def _normalize_resource(resource):
    def norm_item(item):
        item = item.strip()

        if not item:
            return (VOID_PRINCIPAL, FULL_ACCESS_ACTION)

        pair = item.split(ACTION_DELIMETER)

        if len(pair) > 2:
            raise TagthValidationError(f'Too many fields on a resource tag: {item}')

        tag = pair[0]

        if not tag:
            tag = ANYONE_PRINCIPAL

        action = None

        if len(pair) == 2:
            action = pair[1]

        if not action:
            action = FULL_ACCESS_ACTION

        tag = tag.strip()
        action = action.strip()

        if not tag.isidentifier():
            raise TagthValidationError(f'Special characters in resource tag: {tag}')

        if not action.isidentifier():
            raise TagthValidationError(f'Special characters in resource action: {action}')

        return (tag, action)

    if not resource:
        return []

    if not isinstance(resource, str):
        raise TagthValidationError(f'Bad principal {resource}')

    resource = resource.split(TAG_LIST_DELIMETER)
    return list(map(norm_item, resource))


def _resolve_internal(principal, resource):
    actions = set()

    for pr_tag in principal:
        if pr_tag == ROOT_PRINCIPAL:
            actions.add(FULL_ACCESS_ACTION)

    if not resource:
        return actions

    for (res_tag, action) in resource:
        if res_tag == ANYONE_PRINCIPAL:
            actions.add(action)

    for pr_tag in principal:
        if pr_tag == VOID_PRINCIPAL:
            continue

        for (res_tag, action) in resource:
            if res_tag.startswith(pr_tag):
                actions.add(action)

    return actions


def _resolve(principal, resource):
    principal = _normalize_principal(principal)
    resource = _normalize_resource(resource)
    return _resolve_internal(principal, resource)


def allowed(principal, resource, action):
    actions = _resolve(principal, resource)

    if FULL_ACCESS_ACTION in actions:
        return True

    for allowed_action in actions:
        if action.startswith(allowed_action):
            return True

    return False


def validate_principal(principal):
    try:
        _normalize_principal(principal)
    except TagthValidationError:
        return False

    return True


def validate_resource(resource):
    try:
        _normalize_resource(resource)
    except TagthValidationError:
        return False

    return True



from .tagth import (
    allowed,
    validate_principal,
    validate_resource,
    TagthValidationError,
    FULL_ACCESS_ACTION,
    ANYONE_PRINCIPAL,
    ROOT_PRINCIPAL,
    VOID_PRINCIPAL,
    VOID_RESOURCE
)

__all__ = [
    'allowed',
    'validate_principal',
    'validate_resource',
    'TagthValidationError',
    'FULL_ACCESS_ACTION',
    'ANYONE_PRINCIPAL',
    'ROOT_PRINCIPAL',
    'VOID_PRINCIPAL',
    'VOID_RESOURCE'
]



from tagth.tagth import _normalize_principal, _normalize_resource, _resolve, allowed


def test_normalize_principal():
    p = _normalize_principal('me')
    assert list(p) == ['me']

    p = _normalize_principal('this,that, those')
    assert list(p) == ['this', 'that', 'those']


def test_normalize_resource():
    s = _normalize_resource('me')
    assert list(s) == [('me', 'all')]

    s = _normalize_resource('here:admin, there, nowhere:user')
    assert list(s) == [('here', 'admin'), ('there', 'all'), ('nowhere', 'user')]

    s = _normalize_resource('my,:,:ro')
    assert list(s) == [('my', 'all'), ('any', 'all'), ('any', 'ro')]


def test_resolve():
    r = _resolve('me', 'me')
    assert r == {'all'}

    r = _resolve('me', 'meme_me')
    assert r == {'all'}

    r = _resolve('mememe', 'me')
    assert r == set()

    r = _resolve('me', 'mememe:ro, meme:rw')
    assert r == {'ro', 'rw'}

    r = _resolve('me', 'xmememe:ro, meme:rw')
    assert r == {'rw'}


def test_allowed():
    a = allowed('me', 'me', 'all')
    assert a

    a = allowed('me', 'me', 'ro')
    assert a

    a = allowed('me', ':ro', 'ro')
    assert a

    a = allowed('me', 'any:ro', 'ro')
    assert a

    a = allowed('me', 'they:ro, meme:rw', 'rw')
    assert a

    a = allowed('me', 'ther:ro, meme:rw', 'ro')
    assert not a


def test_root():
    a = allowed('a', 'a:ro', 'ro')
    assert a

    a = allowed('a', 'a:ro', 'rw')
    assert not a

    a = allowed('a,root', 'a:ro', 'rw')
    assert a


def test_emply_resource():
    a = allowed('root', '', 'all')
    assert a

    a = allowed('nonroot', '', 'all')
    assert not a


def test_emply_principal():
    a = allowed('', 'resource', 'action')
    assert not a

    a = allowed(',', 'resource', 'action')
    assert not a

    a = allowed(',res', 'resource', 'action')
    assert a

    a = allowed(',res', 'other,resource', 'action')
    assert a

    a = allowed('other,another', 'other,resource', 'action')
    assert a

    a = allowed('other,another', 'another:action', 'action')
    assert a

    a = allowed('other,another', 'another:inaction', 'action')
    assert not a


def test_void():
    a = allowed('resource', 'resource', 'action')
    assert a

    a = allowed('void', 'void', 'action')
    assert not a


def test_empty_anyoune():
    a = allowed('', 'any:action', 'action')
    assert a

    a = allowed('', 'any:action', 'inaction')
    assert not a

    a = allowed('', 'any:all', 'action')
    assert a

    a = allowed('', 'any', 'inaction')
    assert a

    a = allowed('', '', 'action')
    assert not a


def test_normalize_emply():
    n = _normalize_resource('')
    assert not n


def test_action_prefixes():
    assert allowed('user', 'user:a_class_action', 'a_class_action')
    assert allowed('user', 'user:a_class', 'a_class_one')
    assert allowed('user', 'user:a_class', 'a_class_two')
    assert not allowed('user', 'user:b_class', 'a_class_action')
    assert not allowed('user', 'user:a_class', 'b_class_action')
